@@ -12,13 +12,14 @@
 #define TFESC 0xDD
 
 
-
 enum kiss_state_e {
 	KS_SEARCHING,		/* Looking for FEND to start KISS frame. */
-	KS_COLLECTING,		/* In process of collecting KISS frame. */
-	KS_ESCAPE };		/* FESC found in frame. */
+	KS_COLLECTING};		/* In process of collecting KISS frame. */
+
 
 #define MAX_KISS_LEN 2048	/* Spec calls for at least 1024. */
+				/* Might want to make it longer to accomodate */
+				/* maximum packet length. */
 
 #define MAX_NOISE_LEN 100
 
@@ -27,6 +28,8 @@ typedef struct kiss_frame_s {
 	enum kiss_state_e state;
 
 	unsigned char kiss_msg[MAX_KISS_LEN];
+				/* Leading FEND is optional. */
+				/* Contains escapes and ending FEND. */
 	int kiss_len;
 
 	unsigned char noise[MAX_NOISE_LEN];
@@ -36,9 +39,10 @@ typedef struct kiss_frame_s {
 
 
 
-int kiss_frame (kiss_frame_t *kf, unsigned char ch, int debug, void (*sendfun)(int,unsigned char*,int)); 
+int kiss_encapsulate (unsigned char *in, int ilen, unsigned char *out);
+
+void kiss_rec_byte (kiss_frame_t *kf, unsigned char ch, int debug, void (*sendfun)(int,unsigned char*,int)); 
  
-void kiss_process_msg (kiss_frame_t *kf, int debug);
 
 typedef enum fromto_e { FROM_CLIENT=0, TO_CLIENT=1 } fromto_t;
 
